/**
 * Copyright(C) 2013 Patrik Dufresne Service Logiciel <info@patrikdufresne.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *         http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.planimod.core.planif;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.patrikdufresne.ilp.Constraint;
import com.patrikdufresne.ilp.Linear;
import com.patrikdufresne.ilp.LinearProblem;
import com.patrikdufresne.ilp.util.Linears;
import com.patrikdufresne.ilp.util.Variables;
import com.planimod.core.EmployeePreference;
import com.planimod.core.Position;
import com.planimod.core.Task;
import com.planimod.core.Team;

/**
 * This runnable is used to assigned employee to it's preferred position.
 * 
 * @author Patrik Dufresne
 */
@SuppressWarnings("unqualified-field-access")
public class MaxPreferredPosition extends SolveRunnable {

    private static final transient Logger LOGGER = LoggerFactory.getLogger(MaxPreferredPosition.class);

    /** The constraints name. */
    private static final String CONSTRAINT_NAME = "MaxPreferredPosition: SUM from{k in PPi} xik] >= %d, for i=%d"; //$NON-NLS-1$

    /** Constraint to check feasibility */
    private Constraint constraint;

    /** The employee index */
    private int i;

    /** Employee preferred position team */
    private int j;

    /** Employee preferred position */
    private Position pos;

    private boolean released;

    /**
     * When this value is set, we have determine a lower bound using the snapshot value. In case we can't improve this
     * value, the problem will be unfeasible and as such, we must sets this value to the constraint.
     */
    private Number snapshotLowerBound;

    /**
     * Create a new runnable to maximize employee <code>i</code> preferred position.
     * 
     * @param problem
     *            reference to ProblemBuilder
     * 
     * @param i
     *            the employee
     * @param loosly
     *            True to assign the employee to the preferred position event if it's brake the max-senior-employee
     *            rule.
     */
    public MaxPreferredPosition(ProblemBuilder problem, int i) {
        super(problem);
        this.i = i;
        setFeasibilityPumpHeuristic(true);
    }

    private void assignEmployeeToPreferredPosition() throws GeneratePlanifException {
        // For each k, fix the xijk variables to ZERO or ONE if the employee
        // i is assigned to his preferred position.
        for (int k = 0; k < pb.xijk[i][j].length; k++) {
            if (pb.xijk[i][j][k] != null && pb.table.kInRTj(k, j) && ONE.equals(pb.snapshot.get(pb.xijk[i][j][k])) && pos.equals(pb.tasks.get(k).getPosition())) {
                // The employee i is assigned to preferred position k, fix
                // the variables xijk, for each i, and each j.
                for (int j2 = 0; j2 < pb.xijk[i].length; j2++) {
                    if (pb.xijk[i][j2][k] != null) {
                        pb.xijk[i][j2][k].setLowerBound(ONE);
                        // i == i2
                        // && j == j2 ? ONE : ZERO);
                        pb.xijk[i][j2][k].setUpperBound(ONE);
                        // i == i2
                        // && j == j2 ? ONE : ZERO);
                    }
                }
            }
        }
        // Fix to zero (0) xijk and xij variables not corresponding to
        // the team j for employee i. Be careful not to change fixed
        // variables.
        for (int j2 = 0; j2 < pb.xij[this.i].length; j2++) {
            if (j2 != this.j) {
                for (int k = 0; k < pb.xijk[this.i][j2].length; k++) {
                    if (pb.xijk[this.i][j2][k] != null && !Variables.isFixed(pb.xijk[this.i][j2][k])) {
                        pb.xijk[this.i][j2][k].setLowerBound(ZERO);
                        pb.xijk[this.i][j2][k].setUpperBound(ZERO);
                    }
                }
            }
            if (pb.xij[this.i][j2] != null && !Variables.isFixed(pb.xij[i][j2])) {
                pb.xij[this.i][j2].setLowerBound(this.j == j2 ? ONE : ZERO);
                pb.xij[this.i][j2].setUpperBound(this.j == j2 ? ONE : ZERO);
            }
        }
        LOGGER.info("Employee {} assigned to preferred position.", //$NON-NLS-1$
                pb.employeeToString(i));
        // Override the feasible variable.
        this.feasible = true;
    }

    @Override
    protected void feasible() throws GeneratePlanifException {
        assignEmployeeToPreferredPosition();
    }

    @Override
    protected boolean runSolver() throws GeneratePlanifException {
        // The first time, we run the solver without releasing the
        // max-senior-employee rule
        boolean feasible = super.runSolver();
        // Check the results.
        if (feasible || snapshotLowerBound != null && snapshotLowerBound.intValue() > 0) {
            return true;
        }
        // Release the max-senior-employee rule
        this.released = true;
        pb.maxSeniorEmployee.release(i + 1);
        // Try to assign the employee to his preferred position again.
        return super.runSolver();
    }

    @Override
    protected boolean setUp() {
        // Check if the employee has a preferred position
        EmployeePreference pref = pb.preferences.get(i);
        Team team;
        if (pref == null || (this.pos = pref.getPreferredPosition()) == null || (team = pref.getPreferredPositionTeam()) == null) {
            this.feasible = false;
            return false;
        }
        this.j = pb.table.indexOf(team);
        if (j < 0) {
            this.feasible = false;
            return false;
        }
        // TODO To accelerate this, we should sets the xijk not matching the
        // j to ZERO like we do for Preferred Team.
        // Build the linear expression to be maximize.
        Linear linear = pb.lp.createLinear();
        for (int k = 0; k < pb.tasks.size(); k++) {
            Task task = pb.tasks.get(k);
            if (pb.xijk[i][j][k] != null && this.pos.equals(task.getPosition()) && pb.table.kInRTj(k, j) && !ProblemBuilder.isFixedToZero(pb.xijk[i][j][k])) {
                linear.add(pb.lp.createTerm(1, pb.xijk[i][j][k]));
            }
        }
        if (linear.size() == 0) {
            // Noting to maximize, skip the solver
            this.feasible = false;
            return false;
        }
        // Sets the linear as an objective
        pb.lp.setObjectiveLinear(linear);
        pb.lp.setObjectiveDirection(LinearProblem.MAXIMIZE);
        // Also create a constraint to force optimization of the best known
        // solution. (Used to improve performance).
        int lowerBound = Math.max(ZERO.intValue(), (pb.snapshot != null ? (snapshotLowerBound = Linears.compute(linear, pb.snapshot)) : ZERO).intValue());
        String name = String.format(CONSTRAINT_NAME, Integer.valueOf(lowerBound + 1), Integer.valueOf(i));
        constraint = pb.lp.addConstraint(name, linear, Integer.valueOf(lowerBound + 1), null);
        LOGGER.debug("Maximize employee {} preferred position. Optimize: {}", //$NON-NLS-1$
                pb.employeeToString(i),
                lowerBound);
        return true;
    }

    @Override
    protected void tearDown() throws GeneratePlanifException {
        // Release the constraint used to force optimization of the problem.
        if (this.constraint != null) {
            this.constraint.dispose();
            this.constraint = null;
        }
        this.snapshotLowerBound = null;
        // Remove objective
        pb.lp.setObjectiveLinear(null);
        if (this.released) {
            if (feasible) {
                pb.maxSeniorEmployee.run(i + 1);
            } else {
                pb.maxSeniorEmployee.engage(i + 1);
            }
        }
    }

    @Override
    protected void unfeasible() throws GeneratePlanifException {
        if (pos != null && j >= 0) {
            LOGGER.info("Employee {} not assigned to preferred position.", //$NON-NLS-1$
                    pb.employeeToString(i));
        }
    }
}